/*
 *
 */

#include "metric.h"
#include "init.h"
#include "heuristic.h"
#include "conflictOptimal.h"

#define DEBUGRECURSIVE 0

//Global extern variables
extern int nbOfNodes;		//Number of nodes
extern struct node* nodes;	//Array with the nodes
extern int nbOfRadios;		//Number of radios per node
extern int maxFreq;		//Number of orthogonal frequencies


//Compute the conflict graph for a link bidirectional with a conflict at 3 hops.
int conflictGraphLink4(int id1, int id2, int frequency)
{
	int i, j, idN1, idN2, idN3, conflict=0;


	//Si id1 et id2 ne sont pas de voisin
	if(isNeighbor(id1,id2)==0)
	{
	    	return -1;
	}

	// Ne compte pas le lien sans frequence
	if( VIDE==frequency)
	{
		return -1;
	}

	/////////////////////////////////////
	//  Traitement auprès du noeud id1 //
	/////////////////////////////////////
	
	for(i=0;i<nodes[id1].nbN1;i++)
	{
		idN1=nodes[id1].N1[i].id;//index of the neighbor

		//direct neighbor
		if( linkFrequency(id1,idN1,frequency) && (idN1!=id2))
		{
			conflict++;
		}

		//2 1-neighbors together
		for(j=i+1;j<nodes[id1].nbN1;j++)
		{
			if(linkFrequency(idN1,nodes[id1].N1[j].id,frequency)==1)
			{
				conflict++;
			}
		}

		//neighbor with 2-neighbors
		for(j=0;j<nodes[id1].nbN2;j++)
		{
			if(linkFrequency(idN1,nodes[id1].N2[j].id,frequency)==1)
			{
				conflict++;
			}
		}

	}

	//2 2-neighbors together
	for(i=0;i<nodes[id1].nbN2;i++)
	{
		idN2 = nodes[id1].N2[i].id;//index of the neighbor
		for(j=i+1;j<nodes[id1].nbN2;j++)
		{
			if(linkFrequency(idN2,nodes[id1].N2[j].id,frequency)==1)
			{
				conflict++;
			}
		}
	}

	//neighbor with 3-neighbors of node id1
	for(i=0;i<nodes[id1].nbN2;i++)
	{
		idN2 = nodes[id1].N2[i].id;
		for(j = 0 ; j < nodes[idN2].nbN1;j++)
		{
			idN3 = nodes[idN2].N1[j].id;

			if( is3Neighbor(idN3,id1) && (linkFrequency(idN2,idN3,frequency)==1) )
			{
				conflict++;
			}
		}
	}


	/////////////////////////////////////
	//  Traitement auprès du noeud id2 //
	/////////////////////////////////////

	//2 2-neighbors together
	for(i=0;i<nodes[id2].nbN2;i++)
	{
		idN2 = nodes[id2].N2[i].id;//index of the neighbor
		for(j=i+1;j<nodes[id2].nbN2;j++)
		{
			if(linkFrequency(idN2,nodes[id2].N2[j].id,frequency)==1)
			{
				conflict++;
			}
		}
	}

	// 3-neighbors of node id2
	for(i = 0 ; i < nodes[id2].nbN2;i++)
	{
		idN2 = nodes[id2].N2[i].id;
		for(j = 0 ; j < nodes[idN2].nbN1;j++)
		{
			idN3 = nodes[idN2].N1[j].id;

			// idN3 est le voisinage 3-sauts de id2
			if( linkFrequency(idN2,idN3,frequency)==1 && (is3Neighbor(idN3,id2)==1)
			 && (isNeighbor(id1,idN3)==0) && (is2Neighbor(id1,idN3)==0) && (is3Neighbor(id1,idN3)==0))  // et idN3 n'est pas de voisinage 1,2,3 saut de id1
			{
				conflict ++;
			}
		}
	}

	return conflict;
}




void assignN1radioFreq(int nodeID1, int nodeID2, int freq) 
{
	int index, radio;

	//Update radio and freq in the N1 array - the frequencies were allocated in the frequency array only
	//It is this array N1 that is used to update the static route in NS-3
	index=getIndexN1(nodes[nodeID1],nodeID2);
	radio=getRadioByFreq(nodes[nodeID1],freq);
	nodes[nodeID1].N1[index].frequency=freq;
	nodes[nodeID1].N1[index].radio=radio;

	index=getIndexN1(nodes[nodeID2],nodeID1);
	radio=getRadioByFreq(nodes[nodeID2],freq);
	nodes[nodeID2].N1[index].frequency=freq;
	nodes[nodeID2].N1[index].radio=radio;

}

//On teste avec les noeuds d'indices inferieures a id uniquement 
int verifieContrainteVoisinage(int id)
{
	int j, i, idNgh;

	//Verification des contraintes : tous les noeuds doivent avoir une freq commune avec leur voisins
	
	for(j=0;j<id;j++)
	{
		for(i=0;i<nodes[j].nbN1;i++)
		{
			idNgh=nodes[j].N1[i].id;
			if(linkExist2(j,idNgh)==0) return(0);
		}
	}

	return(1);
}


int verifieContrainteLocale(int id)
{
	int i, idNgh;

	//Verification des contraintes : tous les voisins de id et d'indices < id doivent avoir une freq commune avec id
	for(i=0;i<nodes[id].nbN1;i++)
	{
		idNgh=nodes[id].N1[i].id;
		if( (idNgh<id) && (linkExist2(id,idNgh)==0) ) return(0);
	}

	return(1);
}

//Compute the conflict graph
int computeConflictGraph()
{
	int conflict = 0;
	int id, j, id2, freq;

	for(id=0;id<nbOfNodes;id++)
	{
    		for(j=0;j<nodes[id].nbN1;j++)
    		{
			id2=nodes[id].N1[j].id;

			//On parcours l'ensemble des frequences communes aux deux noeuds,
			//car on compte tous les conflits (deux noeuds peuvent avoir plusieurs liens communs).

			for(freq=0; freq<maxFreq;freq++)
			{
				if( (usedfreq(id,freq)!=-1) && (usedfreq(id2,freq)!=-1) )
				{
					conflict += conflictGraphLink4 (id,id2,freq);
				}
			}

    		}//for id2
	}//for id

	conflict = conflict / 2;

	return conflict;
}

//Fonction recursive de parcours de toutes les possibilites
//Return the bestConflict value.
int recursiveTraversal(int nodeID, int bestConflict, struct node* optimalAlloc) 
{
	int radio, freq, conflictNumber, max;

	if(DEBUGRECURSIVE) fprintf(stderr,"--DEBUG recursivetraversal()-- Node %d bestConflict=%d\n",nodeID,bestConflict);

	for(radio=0;radio<nbOfRadios;radio++)
	{
		//We start from -1 (empty radio). It allows to allocate only a part of the radios leading to the minimal conflict graph. 
		//We limit the number of combination in considering that freq on the radios are ordered (radio 0 has the max freq, etc.)
		//Radios have different frequency except for -1 (several radios may be off). 
		if(radio==0) max=maxFreq;
		else max=nodes[nodeID].frequency[radio-1];

		for(freq=-1;freq<max;freq++)
		{
			if( ((nodeID==0) || (nodeID==1)) && (DEBUGRECURSIVE) ) fprintf(stderr,"--DEBUG recursivetraversal()-- Node %d radio=%d freq=%d\n",nodeID,radio,freq);

			//Frequency assignation
			nodes[nodeID].frequency[radio] = freq;
			if(DEBUGRECURSIVE) fprintf(stderr," \t -- Node %d assign freq %d on radio %d\n",nodeID,freq,radio);

			//Constraint Verification 
			if(verifieContrainteLocale(nodeID))
			{ 
				//It is the last node 
				if (nodeID==nbOfNodes-1) 
				{
					conflictNumber=computeConflictGraph();
					if(DEBUGRECURSIVE) fprintf(stderr," \t -- Last node (%d) conflictNumber for this asignation= %d\n",nodeID,conflictNumber);
					if (conflictNumber < bestConflict)
					{
						//We find an optimal (maybe local)  -> we save 
						if(DEBUGRECURSIVE) fprintf(stderr," \t -- is better we keep old=%d new=%d\n",bestConflict,conflictNumber);
						bestConflict=conflictNumber;
						memcpy(optimalAlloc,nodes,nbOfNodes*sizeof(struct node));
					}
				} else {
					//We are not on the last node: we don't compute the conflict --> it is the recursive part
					if(DEBUGRECURSIVE) fprintf(stderr," \t -- We are not on the last node (%d) --> next\n",nodeID);
					bestConflict=recursiveTraversal(nodeID+1, bestConflict, optimalAlloc);
				}
			}//end of if(verifieConstraint()) --> there is an implicit continue in the case where the constraints are not verified

		}
	}//for radio

	return(bestConflict);
}//end of fct recursiveTraversal

//Allocate the frequency to the nodes
float allocateConflictOptimal(int nbOfRoutes)
{
	struct node* optimalAlloc;
	int id, radio, bestConflict, optimal;

	//Copie de sauvegarde hebergeant l'optimal 
	if((optimalAlloc=(struct node *) malloc(nbOfNodes*sizeof(struct node)))==NULL) error("Error: main() -> malloc nodes");

	//Initial value of the number of conflict (the max possible)
	bestConflict=nbOfNodes*nbOfNodes*nbOfNodes*nbOfNodes; //maxNbOfLinks * maxNbOfLinks

	//On assigne une frequence commune (mais a jour aussi les couples radio et interface a utiliser pour atteindre un voisin).
	//Important car tous les liens ne seront pas visites par l'algo

	//On assigne toutes les radios de tous les noeuds sur la freq 0
	//Necessaire pour la verification des contraintes locales (le point de depart doit verifier les contraintes pour l'algo recursif
	//Fait uniquement pour la radio 0 --> -1 sur les autres
   	for(id=0;id<nbOfNodes;id++)
	{
		nodes[id].frequency[0]=0;

   		for(radio=1;radio<nbOfRadios;radio++)
		{
			//nodes[id].frequency[radio]=-1;
			nodes[id].frequency[radio]=0;
		}
	}

	//struct node* nodes is the working list (that is modified to find the optimal)
	//We copy the initial value in optimalAlloc that will host the best allocation.
	memcpy(optimalAlloc,nodes,nbOfNodes*sizeof(struct node));

	//Remarque importante: tous les liens existeront (ce qui n'est pas le cas pour notre algo) ce qui peut diminuer la capacite (du a cette contrainte)

	//The algorithm
	bestConflict=recursiveTraversal(0, bestConflict, optimalAlloc);
	if(DEBUGRECURSIVE) fprintf(stderr,"--DEBUG-- Best conflict=%d \n",bestConflict);

	//Copy the best in the nodes array
	memcpy(nodes,optimalAlloc,nbOfNodes*sizeof(struct node));

	//update the radios/freq in the nodes structure
	optimal=updateNodesAfterAllocation(nbOfRoutes);

	//DEBUG
	printArrayNodes();

 	//affichageReseau();
	free(optimalAlloc);

	return(optimal);

}//End of the function allocateFrequencyGreedy();

float updateNodesAfterAllocation(int nbOfRoutes)
{
	int j, src, next, destination, freq, bestFreq, conflict;
	float bestBandwidth, optimal=0.0;


	//Update the routes
 	for(j=0;j<nbOfRoutes;j++)
	{
		if(nodes[0].routingTable[j].destination>=0 && nodes[0].routingTable[j].nextHop>=0)
         	{
			destination=nodes[0].routingTable[0].destination;
			src=0; next=nodes[0].routingTable[j].nextHop;

			while(src!=destination)
			{
				//Recherche d'une freq commune entre src et next
				//Choix de la freq qui maximise la bande passante du lien / conflits
				bestFreq=-1; bestBandwidth=-1; 
				for(freq=0;freq<maxFreq;freq++)
				{
					if( (usedfreq(src,freq)!=-1) && (usedfreq(next,freq)!=-1) ) 
					{
						conflict = conflictGraphLink4 (src,next,freq);
						if( ((1.0*getMetric(src,next)) / (1.0*conflict+1.0)) > bestBandwidth) 
						{
							if(DEBUGRECURSIVE) fprintf(stderr,"--DEBUG-- link %d <-> %d conflict=%d bandwidth=%f oldValue=%f (freq=%d) we change\n",src,next,conflict+1,1.0*getMetric(src,next)/(1.0*conflict+1.0),bestBandwidth,freq);
							bestBandwidth=1.0*getMetric(src,next)/(1.0*conflict+1.0);
							bestFreq=freq;
						}
					}
				}

				if(bestFreq==-1)
				{
					fprintf(stderr,"Error updateNodesAfterAllocation() - No common frequency between two nodes on a route\n");
					fprintf(stderr,"--DEBUG-- Node %d radio: %d %d -  Node %d radio: %d %d\n",src,nodes[src].frequency[0],nodes[src].frequency[1],next,nodes[next].frequency[0],nodes[next].frequency[1]);

					exit(3);
				}

				//Update radio and freq in the N1 array - the frequencies were allocated in the frequency array only
				//It is this array N1 that is used to update the static route in NS-3
				assignN1radioFreq(src, next, bestFreq);

				src=next;  next=nodes[src].routingTable[j].nextHop;
			}
        	}
        	nodes[0].routingTable[j].cost2=calculMinCapaciteChemin(j);
        	optimal+=nodes[0].routingTable[j].cost2;
      	}
	return(optimal);	

}//end of upateNodesAfterAllocation()

