/*
 */

#include "heuristic.h"

#define CHECKLISTSIZE	2000


//Fonction qui trie les liens par capacite
//Le tableau list est mis a jour avec les indices du noeud emetteur sur le chemin
//par exemple pour le chemin 0->1->2->3->4 la liste sera 0 ; 1 ; 2; 3
//renvoi le nombre de liens, donc 4 pour l'exemple ci-dessus. 
int sortLinkCapacity(int list[], int sizeOfList, int pathNumber, int source, int destination)
{
	//Three local arrays: list of the node (their index on the path), their respective capacity, and an array that indicates if the node has been listed (0 if no ; 1 otherwise)
	int *indexArray, *done; 
	float* capacity;
	int nbOfLinks=0, current, next, bestIndex, i, j;
	float min;

	if( (indexArray=(int *) malloc(sizeOfList*sizeof(int)))==NULL) {perror("Error malloc() in sortLinkCapacity()"); exit(2);}
	if( (capacity=(float *) malloc(sizeOfList*sizeof(float)))==NULL) {perror("Error malloc() in sortLinkCapacity()"); exit(2);}
	if( (done=(int *) malloc(sizeOfList*sizeof(int)))==NULL) {perror("Error malloc() in sortLinkCapacity()"); exit(2);}
	
	//On initialise les tableaux
	current=source; next = nodes[source].routingTable[pathNumber].nextHop;
	while( (nbOfLinks<sizeOfList) && (current!=destination) )	
	{
		indexArray[nbOfLinks]=current;
		capacity[nbOfLinks]=getMetric(current,next);
		done[nbOfLinks]=0;

		//Increment
		nbOfLinks++; 
		current=next;
		next = nodes[current].routingTable[pathNumber].nextHop;
	}

	//We sort
	for(i=0;i<nbOfLinks;i++)
	{
		min=0.0; bestIndex=-1; 
		for(j=0;j<nbOfLinks;j++)
		{
			if( (done[j]==0) && (capacity[j]<min || min <=0.0) )
			{
				bestIndex=j; 
				min=capacity[j]; 
			}
		}

		if(bestIndex==-1){ fprintf(stderr,"Error sortLinkCapacity()\n"); exit(3); }

		list[i]=indexArray[bestIndex];
		done[bestIndex]=1;
	}

	free(indexArray); free(capacity); free(done);
	return(nbOfLinks);
}

//indicates if the node id1 is in the 3-neighborhood of node id2
//return 1 if yes and 0 otherwise
int is3Neighbor(int id1, int id2)
{
	int i,k;

	//If id1=id2 or they are neighbor, then they are NOT 3-hop neighbor (return 0)
	if ( (id1==id2) || (isNeighbor(id1,id2)) || (is2Neighbor(id1,id2))) return 0;

	//But, if one of the 2-neighbors of id1 is neighbor with id2, then they are 3-neighbor.
	for (i=0;i<nodes[id2].nbN2;i++)
	{
		k=nodes[id2].N2[i].id;

		//If we find a common neighbor, they are 2-hop neighbor, return 1
		if (isNeighbor(id1,k)){
			return 1;
		}
	}

	return 0;
}

//Return 1 if the frequency "frequency" is used between the nodes "emetteur" and "recepteur" for an active route "emetteur -> recepteur"
//Otherwise, returns 0.
//Attention: it means that if the route is active for "recepteur->emetteur" with a valid frequency, the function returns 0
int linkFrequencyRouteDirectionnel(int emetteur, int recepteur, int frequency)
{
	// Verifier que le noeud emetteur et recepteur utilisent la frequence 'frequency' en commun
	// et teste si le lien "emetteur->recepteur" est actif et PAS le lien "recepteur -> emetteur"
	return ( linkUsedFrequency(emetteur, recepteur, frequency) && linkIsActive(emetteur,recepteur) && (!linkIsActive(recepteur,emetteur)) ); 
}

//Return 1 if the frequency "frequency" is used between the nodes "emetteur" and "recepteur" for an active route "emetteur -> recepteur"
// and an active route "recepteur-> emetteur".
int linkFrequencyRouteBidirectionnel(int emetteur, int recepteur, int frequency)
{
	// Verifie que le noeud emetteur et recepteur utilisent la frequence 'frequency' en commun
	// et teste si le lien a une route dans les deux sens: "emetteur -> recepteur" et "recepteur --> emetteur" 
	return ( linkUsedFrequency(emetteur, recepteur, frequency) && linkIsActive(emetteur,recepteur) && linkIsActive(recepteur,emetteur) ); 
}


int updateCheckList(int** checkList, int id1, int id2, int checkListNumber)
{
	if(checkListNumber>=CHECKLISTSIZE)
	{
		fprintf(stderr,"Error in updateCheckList(): size of the array CHECKLISTNUMBER is not sufficiently great (check also in costFunction.c ; there is a define)\n");
		exit(1);
	}

	if(id1<id2){ checkList[0][checkListNumber]=id1; checkList[1][checkListNumber]=id2; }
	else{ checkList[0][checkListNumber]=id2; checkList[1][checkListNumber]=id1; }

	return(checkListNumber+1);
}

int isIntheCheckedList(int** checkList, int id1, int id2, int checkListNumber)
{
	int i, found=0;

	for(i=0;i<checkListNumber;i++)
	{
		if(id1<id2 && id1==checkList[0][i] && checkList[1][i]==id2) found=1; 
		if(id1>=id2 && id2==checkList[0][i] && checkList[1][i]==id1) found=1; 
	}

	return(found);
}

//Compute the conflict graph for a link bidirectional with a conflict at 3 hops.
int conflictGraphLink2(int id1, int id2, int frequency)
{
    int i, j, idN1, idN2, idN3, conflict=0;
    int* checkList[2];//Keep all links already checked/counted in the conflict graph
    int checkListNumber=0;


    	// si id1 et id2 ne sont pas de voisin
    	if(isNeighbor(id1,id2)==0) return -1;

    	//Ne compte pas le lien sans frequence
    	if( VIDE==frequency) return -1;

    	if( (checkList[0]=(int *) malloc(CHECKLISTSIZE*sizeof(int)))==NULL) {perror("Error malloc() in conflictGraphLink2()"); exit(2);}
    	if( (checkList[1]=(int *) malloc(CHECKLISTSIZE*sizeof(int)))==NULL) {perror("Error malloc() in conflictGraphLink2()"); exit(2);}

    	/////////////////////////////////////
    	//  Traitement auprès du noeud id1 //
    	/////////////////////////////////////
	for(i=0;i<nodes[id1].nbN1;i++)
	{
		idN1=nodes[id1].N1[i].id;//index of the neighbor

		//direct neighbor
		if( linkFrequencyRoute(id1,idN1,frequency) && (idN1!=id2)){
			checkListNumber=updateCheckList((int **) checkList,id1,idN1,checkListNumber);
			conflict++;
		}

		//2 1-neighbors together
		for(j=i+1;j<nodes[id1].nbN1;j++){
			if(linkFrequencyRoute(idN1,nodes[id1].N1[j].id,frequency)==1){
				checkListNumber=updateCheckList((int **) checkList,idN1,nodes[id1].N1[j].id,checkListNumber);
				conflict++;
			}
		}

		//neighbor with 2-neighbors
		for(j=0;j<nodes[id1].nbN2;j++){
			if(linkFrequencyRoute(idN1,nodes[id1].N2[j].id,frequency)==1){
				checkListNumber=updateCheckList((int **) checkList,idN1,nodes[id1].N2[j].id,checkListNumber);
				conflict++;
			}
		}

	}

	//2 2-neighbors together
	for(i=0;i<nodes[id1].nbN2;i++)
	{
		idN2 = nodes[id1].N2[i].id;//index of the neighbor
		for(j=i+1;j<nodes[id1].nbN2;j++)
		{
			if(linkFrequencyRoute(idN2,nodes[id1].N2[j].id,frequency)==1)
			{
				checkListNumber=updateCheckList((int **) checkList,idN2,nodes[id1].N2[j].id,checkListNumber);
				conflict++;
			}
		}
	}

	//2-neighbor with 3-neighbors of node id1
	for(i=0;i<nodes[id1].nbN2;i++)
	{
		idN2 = nodes[id1].N2[i].id;
		for(j = 0 ; j < nodes[idN2].nbN1;j++)
		{
			idN3 = nodes[idN2].N1[j].id;

			if( is3Neighbor(idN3,id1) && (linkFrequencyRoute(idN2,idN3,frequency)==1) )
			{
				checkListNumber=updateCheckList((int **) checkList,idN2,idN3,checkListNumber);
				conflict++;
			}
		}
	}


	   /////////////////////////////////////
	   //  Traitement auprès du noeud id2 //
	   /////////////////////////////////////

		//2 2-neighbors together
		for(i=0;i<nodes[id2].nbN2;i++)
		{
			idN2 = nodes[id2].N2[i].id;//index of the neighbor
			for(j=i+1;j<nodes[id2].nbN2;j++)
			{
				if( isIntheCheckedList((int**) checkList, idN2,nodes[id2].N2[j].id, checkListNumber)==0 && linkFrequencyRoute(idN2,nodes[id2].N2[j].id,frequency)==1)
				{
					checkListNumber=updateCheckList((int **) checkList,idN2,nodes[id2].N2[j].id,checkListNumber);
					conflict++;
				}
			}
		}

      		// 3-neighbors of node id2
		for(i = 0 ; i < nodes[id2].nbN2;i++)
		{
			idN2 = nodes[id2].N2[i].id;
			for(j = 0 ; j < nodes[idN2].nbN1;j++)
			{
				idN3 = nodes[idN2].N1[j].id;

				// idN3 est le voisinage 3-sauts de id2
				if( isIntheCheckedList((int**) checkList, idN2,idN3, checkListNumber)==0 && linkFrequencyRoute(idN2,idN3,frequency)==1 && (is3Neighbor(idN3,id2)==1)
				 && (isNeighbor(id1,idN3)==0) && (is2Neighbor(id1,idN3)==0) && (is3Neighbor(id1,idN3)==0))  // et idN3 n'est pas de voisinage 1,2,3 saut de id1
				{
					checkListNumber=updateCheckList((int **) checkList,idN2,idN3,checkListNumber);
					conflict ++;
				}
			}
		}



	free(checkList[0]); free(checkList[1]);
	return conflict;
}

//Compute the conflict graph for a link directional with a conflict at 3 hops.
//The only differnce is:
// 0 --> 1 -- 2 -- 3 <-- 4 where links (0,1) and (3,4) are not interfering/conflicting
int conflictGraphLink3(int id1, int id2, int frequency)
{
    	int i, j, idN1, idN2, idN3, conflict=0;
    	int* checkList[2];//Keep all links already checked/counted in the conflict graph
    	int checkListNumber=0, emetteur=-1, recepteur=-1;

    	//Si id1 et id2 ne sont pas de voisin
    	if(isNeighbor(id1,id2)==0) return -1;

	//On ne compte pas le lien sans frequence
	if( VIDE==frequency) return -1;

	//Si le lien i(id1,id2) est bidirectionnel alors on calcul le conflit bidirectionnel
	if(linkFrequencyRouteBidirectionnel(id1, id2, frequency)) return(conflictGraphLink2(id1, id2, frequency));


	//Le lien est directionnel - on fixe emetteur et recepteur
	if(linkFrequencyRouteDirectionnel(id1, id2, frequency)){ emetteur=id1; recepteur=id2;} 
	if(linkFrequencyRouteDirectionnel(id2, id1, frequency)){ emetteur=id2; recepteur=id1;} 

	//Voir si on doit traiter le cas ou il n'y pas de routes entre id1 et id2
	if( (!linkFrequencyRouteDirectionnel(id1, id2, frequency)) && (!linkFrequencyRouteDirectionnel(id2, id1, frequency)) )
	{
		fprintf(stderr,"Error conflictGraphLink3(): no active routes on the link.\n");
		return(-1);
	}

	//Allocation du tableau permettant de s'assurer que chaque lien n'est compte qu'une seule fois
    	if( (checkList[0]=(int *) malloc(CHECKLISTSIZE*sizeof(int)))==NULL) {perror("Error malloc() in conflictGraphLink2()"); exit(2);}
    	if( (checkList[1]=(int *) malloc(CHECKLISTSIZE*sizeof(int)))==NULL) {perror("Error malloc() in conflictGraphLink2()"); exit(2);}

	//Algorithme: 
	// traitement cote Recepteur d'abord:
	//       - les liens avec N1 - entre N1 - entre N1-N2 - et N2-N2 entre eux sont comptes comme conlifts 
	//       - les liens entre N2 et N3 (au sens strict) ne sont comptes comme conflit que de N2-->N3 et pas de N3-->N2 
	// traitement cote recepteur:
	//	 - les liens qui reste sont comptes comme conflits. 

	// recepteur-N1 ; N1-N1 (cote recepteur) ; N1-N2 (cote recepteur)
	for(i=0;i<nodes[recepteur].nbN1;i++)
	{
		idN1=nodes[recepteur].N1[i].id;//index of the neighbor

		//direct neighbor
		if( linkFrequencyRoute(recepteur,idN1,frequency) && (idN1!=emetteur)){
			checkListNumber=updateCheckList((int **) checkList,recepteur,idN1,checkListNumber);
			conflict++;
		}

		//1-neighbors together
		for(j=i+1;j<nodes[recepteur].nbN1;j++){
			if(linkFrequencyRoute(idN1,nodes[recepteur].N1[j].id,frequency)==1){
				checkListNumber=updateCheckList((int **) checkList,idN1,nodes[recepteur].N1[j].id,checkListNumber);
				conflict++;
			}
		}

		//neighbor with 2-neighbors
		for(j=0;j<nodes[recepteur].nbN2;j++){
			if(linkFrequencyRoute(idN1,nodes[recepteur].N2[j].id,frequency)==1){
				checkListNumber=updateCheckList((int **) checkList,idN1,nodes[recepteur].N2[j].id,checkListNumber);
				conflict++;
			}
		}
	}

	//2-neighbors together - cote recepteur
	for(i=0;i<nodes[recepteur].nbN2;i++)
	{
		idN2 = nodes[recepteur].N2[i].id;//index of the neighbor
		for(j=i+1;j<nodes[recepteur].nbN2;j++)
		{
			if(linkFrequencyRoute(idN2,nodes[recepteur].N2[j].id,frequency)==1)
			{
				checkListNumber=updateCheckList((int **) checkList,idN2,nodes[recepteur].N2[j].id,checkListNumber);
				conflict++;
			}
		}
	}

	// emetteur : N1-N2 (cote emetteur)
	for(i=0;i<nodes[emetteur].nbN1;i++)
	{
		idN1=nodes[emetteur].N1[i].id;//index of the neighbor

		//direct neighbor : deja compte (par N1-N2 cote recepteur)
		//1-neighbors together : deja compte (par N2-N2 cote recepteur)

		//neighbor with 2-neighbors
		for(j=0;j<nodes[emetteur].nbN2;j++){
			if( isIntheCheckedList((int**) checkList, idN1,nodes[emetteur].N2[j].id, checkListNumber)==0 && linkFrequencyRoute(idN1,nodes[emetteur].N2[j].id,frequency)==1){
				checkListNumber=updateCheckList((int **) checkList,idN1,nodes[emetteur].N2[j].id,checkListNumber);
				conflict++;
			}
		}
	}

	//2-neighbor with 3-neighbors of node id1 - c'est ici que l'on regarde le sens: on compte pas N3-->N2 mais on compte N2-->N3
	for(i=0;i<nodes[recepteur].nbN2;i++)
	{
		idN2 = nodes[recepteur].N2[i].id;
		for(j = 0 ; j < nodes[idN2].nbN1;j++)
		{
			idN3 = nodes[idN2].N1[j].id;

			//On  doit verifier qu'ils n'ont pas deja ete traite (par N1-N2 cote emetteur)
			if( isIntheCheckedList((int**) checkList, idN2, idN3, checkListNumber)==0 && is3Neighbor(idN3,recepteur) && (linkFrequencyRouteDirectionnel(idN2,idN3,frequency)==1) )
			{
				checkListNumber=updateCheckList((int **) checkList,idN2,idN3,checkListNumber);
				conflict++;
			}
		}
	}


    	/*  	Traitement emetteur	*/

	// 2-neighbors together emetteur
	for(i=0;i<nodes[emetteur].nbN2;i++)
	{
		idN2 = nodes[emetteur].N2[i].id;//index of the neighbor
		for(j=i+1;j<nodes[emetteur].nbN2;j++)
		{
			if( isIntheCheckedList((int**) checkList, idN2,nodes[emetteur].N2[j].id, checkListNumber)==0 && linkFrequencyRoute(idN2,nodes[emetteur].N2[j].id,frequency)==1)
			{
				checkListNumber=updateCheckList((int **) checkList,idN2,nodes[emetteur].N2[j].id,checkListNumber);
				conflict++;
			}
		}
	}

      	// 3-neighbors cote emetteur
	for(i = 0 ; i < nodes[emetteur].nbN2;i++)
	{
		idN2 = nodes[emetteur].N2[i].id;
		for(j = 0 ; j < nodes[idN2].nbN1;j++)
		{
			idN3 = nodes[idN2].N1[j].id;

			// idN3 est le voisinage 3-sauts de id2 
			if( isIntheCheckedList((int**) checkList, idN2,idN3, checkListNumber)==0 && linkFrequencyRoute(idN2,idN3,frequency)==1 && (is3Neighbor(idN3,emetteur)==1) ) 
			{
				checkListNumber=updateCheckList((int **) checkList,idN2,idN3,checkListNumber);
				conflict ++;
			}
		}
	}



	free(checkList[0]); free(checkList[1]);

	return conflict;

}//fin de conflict3 -> conflits directionnels

//Calcul le graphe des conflits dans le cas ou on ne prend en compte que les liens qui ont une route active
int sumConflictOnPaths(int directionnal)
{
	int i, j, freq, sum=0;  

	for(i=0; i<nbOfNodes-1; i++)
	{
		for(j=i+1; j<nbOfNodes; j++)
		{
			for(freq=0; freq<maxFreq; freq++)
			{
				if(linkFrequencyRoute(i,j,freq))
				{
					if(directionnal==0) sum+=conflictGraphLink2(i, j, freq); 
					else sum+=conflictGraphLink3(i, j, freq); 
				}
			}
		}
	}

	return(sum);
}


////////////////////////////////////////////////////////////
// 	Algorithme : l'assignation des fréquences en réseau   //
////////////////////////////////////////////////////////////

//Alloue la meme frequence vide partout (frequence -1) et sur la meme radio (-1) les autres radios ne sont pas alloues
void allocateEmptyFrequency()
{
	int id,radio,j;

	//On assigne toutes les radios de tous les noeuds sur la freq vide
	for(id=0;id<nbOfNodes;id++)
	{
		for(radio=0;radio<nbOfRadios;radio++)
			nodes[id].frequency[radio]=VIDE;

		//On fixe les radios/freq vers les voisins a (0,0)
		for(j=0;j<nodes[id].nbN1;j++)
		{
			nodes[id].N1[j].frequency=VIDE;
			nodes[id].N1[j].radio=VIDE;
		}
	}

}

// Nous calculons le conflict directionnel ou bidirectionnel ici
int conflictGraphGeneric(int vertex1, int vertex2)
{
	// récuperer index
	int idVetex2 = getIndexN1(nodes[vertex1],vertex2);
	int frequence = nodes[vertex1].N1[idVetex2].frequency;

	int IE; 	// IE = nombre de conflict + 1

	// Global variables : directionnels
	// If 1 there is conflict directional
	// If 0 there is conflict bidirectional
	if(directionnels == 0)
	{
		//Compute the conflict graph for a link bidirectional with a conflict at 3 hops.
		IE = conflictGraphLink2(vertex1,vertex2,frequence) + 1 ;
	}
	else if(directionnels == 1)
	{
		//Compute the conflict graph for a link directional with a conflict at 3 hops.
		IE = conflictGraphLink3(vertex1,vertex2,frequence) + 1;
	}

	return IE;
}

// Rechercher la plus petite capacité sur le chemin pathNumber
float calculMinCapaciteChemin(int pathNumber)
{
	int source = 0, conflict;
	int destination = nbOfNodes - 1;
	int next = nodes[source].routingTable[pathNumber].nextHop;

	//Si cette route ne peut pas arriver à la destination, on returne la valeur zéro
	if(nodes[source].routingTable[pathNumber].destination ==-1)
	{
		// sauvegarder la plus petite valeur  B(i)/IE(i) sur le chemin pathNumber
		nodes[source].routingTable[pathNumber].cost2 = 0.0;
		return 0.0;
	}


	// fonction getMetric : prendre le débit sur le lien (source,next)
	// fonction conflictGraphGeneric : calculer le conflict (IE) pour le lien (source,next)

	float min ;
	if(conflictGraphGeneric(source,next) > 0 )
	{
		min =  (1.0 * getMetric(source,next)) / (conflictGraphGeneric(source,next) * 1.0);
	}
	else // dans ce cas, ce lien a la fréquence vide (conflictGraphGeneric(source,next) est égale -1)
	{
		/*
		 *         vide     f1     f2
		 * 		S -------A------B-----D
		 *
		 * */

		min =  0.0;
	}

	//parcourir le chemin pathNumber
	if(next!=destination)
	{
		do
		{
			source = next;
			next = nodes[source].routingTable[pathNumber].nextHop;
			conflict = conflictGraphGeneric(source,next) ;

			if(conflict > 0 )
			{
				float capacite = (1.0 * getMetric(source,next)) / (conflictGraphGeneric(source,next) * 1.0);
				if( capacite < min) min = capacite;		
			} 
		}while(next != destination);
	}

	// sauvegarder la plus petite valeur  B(i)/IE(i) sur le chemin pathNumber
	nodes[source].routingTable[pathNumber].cost2 = min;

	return min;
}

// Calculer la capacité du réseau
float calculCapaciteReseau(int nbOfRoutes)
{
	int i;
	float capacite = 0.0;
	for (i=0 ; i < nbOfRoutes ; i++)
		capacite += calculMinCapaciteChemin(i);

	return capacite;
}

// compter le nombre de 'freq' qui avait utilisé sur le noeud vertex1
int countFrequency(int vertex1, int freq)
{

	if(freq==VIDE)
		return -1;

	int resul = 0 ;
	int i;
	for(i=0;i < nodes[vertex1].nbN1 ; i++)
		if(nodes[vertex1].N1[i].frequency == freq)
			resul ++;

	return resul;
}

// compter le nombre de radio qui avait utilisé sur le noeud vertex1
int countRadio ( int vertex1 )
{
	int resul = 0 ;
	int i;
	for(i=0;i < nbOfRadios ; i++)
		if(nodes[vertex1].frequency[i] != -1)
			resul ++;

	return resul;
}

// Changer la fréquence entre deux noeuds vertex1 et vertex2
// Returner -1, si il est impossible
int updateFreq(int vertex1,int vertex2, int frequence)
{

	// récuperer index
	int idVetex1 = getIndexN1(nodes[vertex2],vertex1);
	int idVetex2 = getIndexN1(nodes[vertex1],vertex2);

	// la fréquence commune entre vertex1 et vertex2
	int freq = nodes[vertex1].N1[idVetex2].frequency;

	// s'il y a déjà cette fréquence
	if(freq == frequence)
		return 1;


	// compter le nombre de 'freq' qui avait utilisé sur le noeud vertex1 et le noeud vertex2
	// Si fre est la fréquence vide, return -1
	int numOldFreqNode1 = countFrequency( vertex1, freq);
	int numOldFreqNode2 = countFrequency( vertex2, freq);

	// compter le nombre de 'frequence' qui avait utilisé sur le noeud vertex1 et le noeud vertex2
	// Si 'frequence' est la fréquence vide, return -1
	int numNewFreqNode1 = countFrequency( vertex1, frequence);
	int numNewFreqNode2 = countFrequency( vertex2, frequence);

	// compter le nombre de radio qui avait utilisé sur le noeud vertex1
	int numRadioNode1 = countRadio ( vertex1 );

	// compter le nombre de radio qui avait utilisé sur le noeud vertex2
	int	numRadioNode2 = countRadio ( vertex2 );

	int passNode1 = numNewFreqNode1 > 0 || numOldFreqNode1 == 1 || numRadioNode1 < nbOfRadios || frequence==VIDE;
	int passNode2 = numNewFreqNode2 > 0 || numOldFreqNode2 == 1 || numRadioNode2 < nbOfRadios || frequence==VIDE;

	int newRadioIndex = -1;

	// Si nous pouvons changer la fréquence
	if (passNode1 && passNode2)
	{

		// Mis à jour la nouvelle fréquence sur le noeud vertex1 et vertex2
		nodes[vertex1].N1[idVetex2].frequency=frequence;
		nodes[vertex2].N1[idVetex1].frequency=frequence;

		// Changement la radio sur le noeud vertex1
		if (numOldFreqNode1 != 1 && numNewFreqNode1 == 0)
		{
			int newRadioNode1 = freeRadio(vertex1); // Chercher la premiere radio disponible sur le noeud 1
			nodes[vertex1].N1[idVetex2].radio = newRadioNode1;
			nodes[vertex1].frequency[newRadioNode1]=frequence;
		}
		else
		{
			if (numNewFreqNode1 == 0)
			{
				// Rechercher newRadioIndex de 'freq' sur le noeud vertex1
				newRadioIndex = usedfreq( vertex1,freq );
			} else
			{
				// Rechercher newRadioIndex de 'frequence' sur le noeud vertex1
				newRadioIndex = usedfreq( vertex1,frequence );

				// Supprimer freq, si nous ne l'utilise plus
				if ( numOldFreqNode1 == 1 )
					nodes[vertex1].frequency[ usedfreq(vertex1,freq) ] = -1;
			}
			// nodes[vertex1].N1[idVetex2].radio <-- newRadioIndex
			nodes[vertex1].N1[idVetex2].radio = newRadioIndex;
			nodes[vertex1].frequency[nodes[vertex1].N1[idVetex2].radio]=frequence;
		}

		// Changement la radio sur le noeud vertex2
		if (numOldFreqNode2 != 1 && numNewFreqNode2 == 0)
		{
			int newRadioNode2 = freeRadio(vertex2); // Chercher la premiere radio disponible sur le noeud 2
			nodes[vertex2].N1[idVetex1].radio = newRadioNode2;
			nodes[vertex2].frequency[newRadioNode2]=frequence;
		}
		else
		{
			if (numNewFreqNode2 == 0)
			{
				// Rechercher newRadioIndex de 'freq' sur le noeud vertex2
				newRadioIndex = usedfreq( vertex2,freq );
			}
			else
			{
				// Rechercher newRadioIndex de 'frequence' sur le noeud vertex2
				newRadioIndex = usedfreq( vertex2,frequence );

				// Supprimer freq, si nous ne l'utilise plus
				if ( numOldFreqNode2 == 1 )
					nodes[vertex2].frequency[ usedfreq(vertex2,freq) ] = -1;
			}
			// nodes[vertex2].N1[idVetex1].radio <-- newRadioIndex
			nodes[vertex2].N1[idVetex1].radio = newRadioIndex;
			nodes[vertex2].frequency[nodes[vertex2].N1[idVetex1].radio]=frequence;
		}

		return 1;
	}

	return -1;
}


// Trouver la fréquence pour le lien (vertex1,vertex2) qui nous donne la plus grande capacité soit supérieur à la capacité du réseau en cours
float assignationLien(int vertex1, int vertex2,int nbOfRoutes,float* result)
{

	int i,nbOfFa,nbOfFs,nbOfFreq;
	float max = 0.0, capacite;

	// récuperer index de vertex2
	int idVetex2 = getIndexN1(nodes[vertex1],vertex2);
	int bestFreq = nodes[vertex1].N1[idVetex2].frequency;


	int Freq[maxFreq]; // la liste de fréquence à parcourir
	for (i=0 ; i < maxFreq ; i++)
		Freq[i] = -1;


	nbOfFa=nbOfFs=nbOfFreq = 0;
	for(i = 0 ; i < nbOfRadios ; i++)
	{
		if(nodes[vertex1].frequency[i] != -1)
			nbOfFs ++;

		if(nodes[vertex2].frequency[i] != -1)
			nbOfFa ++;
	}

//	int cas = 0;
	// Le noeud vertex1 et vertex2 ont encore des radios libres, la liste Freq contient toutes les fréquences
	if((nbOfFs < nbOfRadios) && (nbOfFa < nbOfRadios))
	{
//		cas = 1;
		nbOfFreq = maxFreq;
		for(i = 0 ; i < maxFreq ; i++)
			Freq[i] = i;
	}
	// le noeud vertex2 n'a plus de radio libre, la liste Freq contient toutes les fréquences de noeud vertex2
	else if ((nbOfFs < nbOfRadios) && (nbOfFa == nbOfRadios))
	{
//		cas = 2;
		for(i = 0 ; i < maxFreq ; i++)
		{
			if(usedfreq(vertex2,i) != -1)
			{
				Freq[nbOfFreq] = i;
				nbOfFreq ++;
			}
		}

	}
	// le noeud vertex1 et vertex2 n'ont plus de radio libre, la liste Freq contient toutes les fréquences commumes
	else if((nbOfFs == nbOfRadios) && (nbOfFa == nbOfRadios))
	{
//		cas = 3;
		for(i = 0 ; i < maxFreq ; i++)
		{
			if ((usedfreq(vertex1,i) != -1) && (usedfreq(vertex2,i) != -1))
			{
				Freq[nbOfFreq] = i;
				nbOfFreq ++;
			}
		}
	}
	// le noeud vertex1 n'a plus de radio libre, la liste Freq contient toutes les fréquences de noeud vertex1
	else if((nbOfFs == nbOfRadios) && (nbOfFa < nbOfRadios))
	{
//		cas = 4;
		for(i = 0 ; i < maxFreq ; i++)
		{
			if(usedfreq(vertex1,i) != -1)
			{
				Freq[nbOfFreq] = i;
				nbOfFreq ++;
			}
		}
	}

	// Parcourir la liste de fréquence
	for(i = 0 ; (i < nbOfFreq) && (Freq[i] != -1); i++)
	{
//		int test = Freq[i];
		// Changement la fréquence de lien (vertex1,vertex2) par la fréquence Freq[i]
		if(updateFreq(vertex1,vertex2,Freq[i]) == 1)
		{
			capacite = estimateThroughput(nbOfRoutes, result, 1);//1 vaut pour l'heuristic / 0 si on recherche l'optimal
			fprintf(stderr,"-- DEBUG assignation lien (%d,%d) freq=%d --> capacite=%f\n",vertex1,vertex2,Freq[i],capacite);

			if(capacite > max)
			{
				max = capacite ;
				bestFreq = Freq[i];
			}
		}
	}
	fprintf(stderr,"-- DEBUG assignation lien (%d,%d) freq optimal=%d --> capacite=%f\n",vertex1,vertex2,bestFreq,max);

	updateFreq(vertex1,vertex2,bestFreq);
	return max;
}

// Vérifier que le lien (vertex1,vertex2) exist déjà sur la route pathNumber.
// Si oui, retourner 1.
// Si non, retourner -1
int existLienSurRoute(int vertex1, int vertex2,int pathNumber)
{
	int source = 0;
	int destination = nbOfNodes - 1;
	int next = nodes[source].routingTable[pathNumber].nextHop;

	do
	{
		if ( (source==vertex1 && next==vertex2) || (source==vertex2 && next==vertex1) )
			return 1;

		source = next;
		next = nodes[source].routingTable[pathNumber].nextHop;
	}while (source != destination);

	return -1;
}

// Vérifier que le lien (vertex1,vertex2) appartient juste la route 'pathNumber'
// Si oui, retourner 1.
// Si non, retourner -1
int existJusteLien(int vertex1, int vertex2,int pathNumber, int nbOfRoutes)
{
	int i;

	// Si le lien (vertex1,vertex2) n'appartient pas la route 'pathNumber'
	if(existLienSurRoute(vertex1,vertex2,pathNumber) == -1)
		return -1;

	// Si le lien (vertex1,vertex2) appartient des routes qui sont différentes à la route 'pathNumber'
	for(i = 0 ; i < nbOfRoutes ; i++)
		if( i != pathNumber && existLienSurRoute(vertex1,vertex2,i) == 1)
			return -1;

	return 1;
}

// on remet la fréquence vide aux liens sur le chemin pathNumber
void initialisationChemin(int pathNumber,int nbOfRoutes)
{
	int source = 0;
	int destination = nbOfNodes - 1;
	int next = nodes[source].routingTable[pathNumber].nextHop;


	fprintf(stderr,"La route %d est supprimee\n",pathNumber);
	do
	{
		// Si ce lien appartient juste la route 'pathNumber', nous y mettons la fréquence vide
		if(existJusteLien(source,next,pathNumber,nbOfRoutes) == 1)
		{
			//Assignation la fréquence vide (-1)
			updateFreq(source,next, VIDE);
		}
		//On met la route a -1 ici pour pas qu'elle ne soit pas utilise dans le routage statique
		nodes[source].routingTable[pathNumber].destination=-1;

		source = next;
		next = nodes[source].routingTable[pathNumber].nextHop;

	}while (source != destination);
}

void affichageReseau()
{
	int id,id2;

	fprintf(stderr,"\n");
	for(id=0;id<nbOfNodes;id++)
	{
    		for(id2=0;id2<nodes[id].nbN1;id2++)
		{
    			if( nodes[id].N1[id2].frequency > -1)
    			{
    				fprintf(stderr,"-Assignation link (%d, %d) radio=%d freq=%d Debit:%f IE=%d\n",
						  id,nodes[id].N1[id2].id,nodes[id].N1[id2].radio,
						  nodes[id].N1[id2].frequency,getMetric(id,nodes[id].N1[id2].id),
						  conflictGraphGeneric(id,nodes[id].N1[id2].id));

    			}
		}
	}
	fprintf(stderr,"\n");
}//end of fct affichageReseau()

//Allocate the frequency to the nodes
//nbOfRoutes is number routing between node source and destination
float allocateHeuristic(int nbOfRoutes)
{
	int i, source, destination,next;
	float result[3];
	float cost1=0.0;

	// Initialisation du réseau : aucune fréquence
	allocateEmptyFrequency();

	float maxCapacite = 0.0; // la capacité du réseau en cours
	float maxLocal = 1.0; // la capacité locale

	for(i = 0 ; i < nbOfRoutes; i++) // parcourir la liste de route
	{
		source = 0;
		destination = nbOfNodes - 1;
		next = nodes[source].routingTable[i].nextHop;

		while ((maxLocal >= maxCapacite) && (source != destination))
		{
			// Trouver la fréquence pour le lien (vertex1,vertex2)
			// qui nous donne la plus grande capacité soit supérieure à maxCapacite
			maxLocal = assignationLien(source,next,nbOfRoutes,result);

//			affichageReseau();

			source = next;
			next = nodes[source].routingTable[i].nextHop;
		}

//		 affichageReseau();

		if(maxLocal >= maxCapacite) // Si la capacité du réseau est augmentée, on met à jour sa valeur
		{
			maxCapacite = maxLocal;
			maxLocal ++;
		}
		else // on remet la fréquence vide aux liens sur ce chemin
		{
			initialisationChemin(i,nbOfRoutes);
		}

	}

	maxCapacite=estimateThroughput(nbOfRoutes, result, 1);

	// sauvegarder la valeur B(i)/IE(i) dans cost2; sert a fixer le debit de la source
	for(i = 0 ; i < nbOfRoutes ; i++) // parcourir la liste de route
		if(nodes[0].routingTable[i].destination>=0)
		{
			nodes[0].routingTable[i].cost2 = result[i];
			cost1 += calculMinCapaciteChemin(i);
		}

	fprintf(stderr,"maxCapacite(heuristic)=%f\n",cost1);
	fprintf(stderr,"--DEBUG-- numberRouteCalcule=%d\n",nbOfRoutes);
	affichageReseau();
	
	return(cost1);
}//End of the function allocateHeuristic()





